; log.scm
; $Id: log.scm,v 1.13 2006/03/23 14:10:27 inger Exp $

;;------------------------------------------------------------------------------
;; transaction logging
;;------------------------------------------------------------------------------
;; LOG, pointed by *packages-log-top* is a list of transactions
;; TRANSACTION is a cons (command . delta), where
;; COMMAND is a list of strings: (cmd . contents)
;;         CMD ::= "install" | "remove"
;; DELTA is a list of elements like COMMAND, that is,
;;       list of lists of strings and represents actual changes
;;       between previous and current packages state
;; transaction example:
;; (("install" "openssh-clients") (("install" "openssh" "openssh-clients")))

(define *packages-log-top* (make-cell '()))
(define *packages-log-point* (make-cell (cell-ref *packages-log-top*)))

(define (packages-log-pending)
  (map
   (lambda (e)
     (let ((address (string-append "/packages/pending/" e)))
       (cons e (sort (woo-list-names address) string<))))
   (filter
    (lambda (e)
      (member e '("install" "remove" "upgrade")))
    (woo-list-names "/packages/pending"))))


(define *packages-log-pending* (make-cell (packages-log-pending)))

(define (packages-log-revert-cmd cmd)
  (cond ((string=? cmd "install") "remove")
        ((string=? cmd "remove") "install")
        (else cmd)))

(define (packages-log-revert-command command)
  (cons (packages-log-revert-cmd (car command)) (cdr command)))

(define (packages-log-apply-command command)
  (let ((cmd (car command))
        (contents (cdr command)))
    (map
     (lambda (p)
       (woo-write
        (string-append "/packages/state/" p)
        'state cmd))
     contents)))

(define (packages-log-delta old new)
  (filter-map
   (lambda (cmd)
     (let* ((old (or (assoc-ref old cmd) '()))
            (new (or (assoc-ref new cmd) '()))
            (forward (lset-difference equal? new old))
            (backward (lset-difference equal? old new)))
       (or
        (and (not (null? forward))
             (cons (packages-log-revert-cmd cmd) forward))
        (and (not (null? backward))
             (cons cmd backward)))))
   (lset-union equal?
               (map car old)
               (map car new))))

(define (packages-log-point-up top point)
  (or (and (eq? (cdr top) point) top)
      (packages-log-point-up (cdr top) point)))

;;; reduces log by eliminating complementary commands
(define (packages-log-reduce top)
  (reverse
   (delete-duplicates
    top
    (lambda (x y)
      (let ((x-command (car x))
            (y-command (packages-log-revert-command (car y))))
        (equal? x-command y-command))))))

(define (packages-log-find top command)
  (or (and (or (null? top)
               (list= equal? command (caar top)))
           top)
      (packages-log-find (cdr top) command)))

;;------------------------------------------------------------------------------

(define (packages-log-add command)
  (let* ((pending (packages-log-pending))
         (delta (packages-log-delta (cell-ref *packages-log-pending*) pending)))
    (cell-set! *packages-log-pending* pending)
    (cell-set! *packages-log-point* (cons (cons command delta)
                                          (cell-ref *packages-log-point*)))
    (cell-set! *packages-log-top* (cell-ref *packages-log-point*))))

(define (packages-log-undo)
  (or (packages-log-undo-empty?)
      (begin
        (map packages-log-apply-command
             (cdar (cell-ref *packages-log-point*)))
        (cell-set! *packages-log-point* (cdr (cell-ref *packages-log-point*)))
        (cell-set! *packages-log-pending* (packages-log-pending)))))

;;; reduces log, then finds complementary command;
;;; if found, replays log, excluding complementary command
;;; oterwise noop
(define (packages-log-undo-trans command)
  (let* ((rcommand (packages-log-revert-command command))
         (top (packages-log-reduce (cell-ref *packages-log-point*)))
         (point (packages-log-find top rcommand)))
    (if (null? point) #f
        (let ((rcommand (caar point)))
          (woo-write "/packages/control" 'control "quit")
          (map
           (lambda (command)
             (or (eq? rcommand command)
                 (packages-log-apply-command command)))
           (map car top))))))

(define (packages-log-redo)
  (or (packages-log-redo-empty?)
      (begin
        (cell-set! *packages-log-point*
                   (packages-log-point-up (cell-ref *packages-log-top*)
                                          (cell-ref *packages-log-point*)))
        (packages-log-apply-command (caar (cell-ref *packages-log-point*)))
        (cell-set! *packages-log-pending* (packages-log-pending)))))

(define (packages-log-set top)
  (cell-set! *packages-log-top* top)
  (cell-set! *packages-log-point* (cell-ref *packages-log-top*))
  (cell-set! *packages-log-pending* (packages-log-pending)))

(define (packages-log-get)
  (cell-ref *packages-log-top*))

(define (packages-log-reset)
  (cell-set! *packages-log-top* '())
  (cell-set! *packages-log-point* (cell-ref *packages-log-top*))
  (cell-set! *packages-log-pending* (packages-log-pending)))

(define (packages-log-undo-empty?)
  (null? (cell-ref *packages-log-point*)))

(define (packages-log-redo-empty?)
  (or (null? (cell-ref *packages-log-top*))
      (eq? (cell-ref *packages-log-point*) (cell-ref *packages-log-top*))))

(define (packages-log-pending-empty?)
  (null? (cell-ref *packages-log-pending*)))
